from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

CANARY_TYPE_BASE: CanaryType
CANARY_TYPE_NEW: CanaryType
CANARY_TYPE_UNSPECIFIED: CanaryType
DESCRIPTOR: _descriptor.FileDescriptor

class CanaryType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
